# ABOSS, filtering Ig-seq data with structural information

ABOSS is a structurally motivated software to filter antibody sequences that might not fold into a viable protein. As such it is an error filtering protocol for Ig-seq outputs.


##Quickstart.

If you simply want to check if the program runs correctly on your system after installation, go to the **Format_examples** directory and run the following command.

```
ABOSS -f Igblastn_example.txt -o Igblastn -n 4 -r 1000
```

If the program finished successfully, it points to a successful installation and you can read below to understand arguments, parameters, inputs and outputs of the program.

## This is the ABOSS manual that will help you to run ABOSS on your machine.

Examples of ABOSS input files are shown in the Format_examples directory, which are **Igblastn outputs**, **FASTA format** and **pickle format**. The fasta and pickle file examples come from Galson et al., (2016) study, whilst Igblastn outputs are from Halliley et al., 2015 study.  ALL sequences are heavy chains that come from human B cell donors.

# Igblastn as input

Make sure that you specified nucleotide translation into amino acids when you run Igblastn.

# Fasta as input
 >line1
 QVQLQQSGAEVVRSGASVKLSCTASGFNIKDYYIHWVKQRPEKGLEWIGWIDPEIGDTEYVPKFQGKATMTADTSSNTAYLQLSSLTSEDTAVYYCNAGHDYAARGRFPYWGQGTLVTVSA

# Pickle format

python dictionary in the **pickle format**, where the key is the amino acid sequence and its value is redundancy.

For example: {"QVQLQQSGAEVVRSGASVKLSCTASGFNIKDYYIHWVKQRPEKGLEWIGWIDPEIGDTEYVPKFQGKATMTADTSSNTAYLQLSSLTSEDTAVYYCNAGHDYAARGRFPYWGQGTLVTVSA": 3}

## COMMAND LINE OPTIONS

#### To run ABOSS:
	
	ABOSS -f [INPUT FILE]
	
	"-h",  <Help message>	
	
	"-f",  <Igblastn, FASTA or PICKLE file of python dictionary>
	
	"-o",  <Specify format from [Igblastn,Fasta,Pickle]>	

	"-c",  <Indicate the variable chain [H or L] of antibodies in the Ig-seq data. H is default>

	"-r",  <Indicate the rate of ABOSS analysis, e.g. how many sequences are parsed with ANARCI at a time. Default rate is 1000>

	"-s",  <Indicate the species [human, mouse, rhesus, rabbit, alpaca] of the Ig-seq data. The default species is human>

	"-n",  <The number of cores engaged in ABOSS analysis. The default number of cores is 8>

## Some examples how to run ABOSS:

The ABOSS analysis of a FASTA file from human V-chain Ig-seq data on a 8-core desktop machine:

	ABOSS -f YOUR_file.fa -o Fasta -c H -r 1000 -s human -n 8

The ABOSS analysis of an Igblasn file from human V-chain Ig-seq on a 44-core server:

	ABOSS -f YOUR_file.txt -o Igblastn -n 44 -r 1000 -c H

The ABOSS analysis of a pickle file from human V-chain Ig-seq on a 8-core desktop machine:

	ABOSS -f YOUR_file.pkl -o Pickle -c H -r 10000 -s human -n 8

##  ABOSS output files

	"Final_Ig-seq_information.csv" - <The CSV file that contains ANARCI Parsed sequences, Redundancy, CDR-H3, Number of flagged residue/positions, Flagged residue/positions (None is no flagged positions), and V and J genes, ANARCI numbered sequences and Query names for Igblastn inputs >

	"Graphs/" - <The folder that contains figures generated in ABOSS analysis. These plots include positional amino acid distributions, Shannon entropy and number of entries in Ig-seq sequences with zero flagged residue/positions as well as the plot of the number of flags in an individual sequence against its redundancy.>

	"Error/ig_seq_residue_error_rate.txt" - <The text file with the calculated residue error rate for your Ig-seq data.>

