import pandas as pd
import cPickle as pickle
from anarci import run_anarci
from optparse import OptionParser
import os
import multiprocessing as mp
import time
from os.path import join
import csv
import json,sys

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))
from region_definitions import Accept

csv.field_size_limit(sys.maxsize)


###################################################
### Accept function to number antibody residues ###
###################################################
a = Accept()
a.numbering_scheme = "imgt"
a.definition = "imgt"

def merge_dicts(*dict_args):
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

def rabbit_length_check(position_73,rabbit_position_84,regions,quality):
	if len(regions["fwh3"]) - position_73 + rabbit_position_84  < 38:   # Accounting for insertions at postion 73 and deletion at 84
		quality = ("Bad", "FW3 length is shorter IMGT defined"), rabbit_position_84
	if len(regions["fwh2"]) < 17:
		quality = ("Bad", "FW2 length is shorter IMGT defined")
	if len(regions["fwh4"]) < 10:
		quality = ("Bad", "FW4 length is shorter IMGT defined")
	if len(regions["cdrh3"]) > 37:   
		quality = ("Bad", "CDR-H3 is chimeric")
	return quality

def standard_species_length_check(position_73,regions,quality):
	if len(regions["fwh3"]) - position_73 < 38:   # Accounting for insertions at postion 73
		quality = ("Bad", "FW3 length is shorter IMGT defined")
	if len(regions["fwh2"]) < 17:
		quality = ("Bad", "FW2 length is shorter IMGT defined")
	if len(regions["fwh4"]) < 11:
		quality = ("Bad", "FW4 length is shorter IMGT defined")
	if len(regions["cdrh3"]) > 37:   
		quality = ("Bad", "CDR-H3 is chimeric")
	return quality

def light_chain_species_length_check(absent_positions,position_73,regions,quality):
	if len(regions["fwl3"]) + absent_positions - position_73 < 38: # Accounting for insertions at postion 73,81,83
		quality = ("Bad", "FW3 length is shorter IMGT defined")
	if len(regions["fwl2"]) < 17:
		quality = ("Bad", "FW2 length is shorter IMGT defined")
	if len(regions["fwl4"]) < 10:
		quality = ("Bad", "FW4 length is shorter IMGT defined")
	return quality

def standard_species_length_check_fw1(fw,all_region_aa,quality):
	frame1 =  [int(x) for x in all_region_aa[fw]] 
	if 10 not in frame1:
		if min(frame1) < 10:
			if max(frame1) - min(frame1) != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")
		else:
			if max(frame1) - min(frame1) + 1 != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")
	else:
		if max(frame1) - min(frame1) + 1 != len(frame1):
			quality = ("Bad", "FW1 length is shorter IMGT defined")
	return quality

def rabbit_length_check_fw1(fw,all_region_aa,rabbit_position_2,quality):
	"""
	Framework1 sometimes misses position 2 in rabbits.
	"""
	frame1 =  [int(x) for x in all_region_aa[fw]]
	if 10 not in frame1:
		if min(frame1) == 1:
			if max(frame1) - min(frame1) - rabbit_position_2 != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")				
		elif min(frame1) >= 3 and min(frame1) < 10:
			if max(frame1) - min(frame1) != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")	
		else:
			if max(frame1) - min(frame1) + 1 != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")
	else:
		if max(frame1) - min(frame1) + 1 - rabbit_position_2 != len(frame1):
			quality = ("Bad", "FW1 length is shorter IMGT defined")
	return quality

def checking_structural_viability(numbering):
	"""
	The main function of ANARCI parsing. Here we check for unusual
	indels in antibody regions across different species.
	We also check if the first numbered position of the amino acid 
	sequence starts before postion 23
	"""
	good_aminos = sorted(list("QWERTYIPASDFGHKLCVNM"))
	all_region_aa = {}
	regions = {}
	quality = "good"
	Conserved_position = 0
	absent_positions = 2
	position_73 = 0
	if species == "rabbit":
		rabbit_position_2 = 1
		rabbit_position_84 = 1
	for frame in list_of_frames:
		all_region_aa[frame] = {}
		region = ''
		a.set_regions([frame])
		for n in range(len(numbering)):
			if a.accept(numbering[n][0], chain) == 1:       # current chain
				if numbering[n][1] not in good_aminos and numbering[n][1] != "-":  # Checking here for unusual aa while allowing for "-" entry at some position (cases of CDRs)
					quality = ("Bad", "Weird amino acid")
				if numbering[n][1] == '-':
						continue
				else:
					if numbering[n][0][1] != ' ':
						all_region_aa[frame][str(numbering[n][0][0]) + numbering[n][0][1]] = numbering[n][1]
						#########################################################
						### Check for indels in frameworks and canonical CDRs ###
						#########################################################
						if chain == "H":
							if frame != "cdrh3":    
								quality = ("bad", str(numbering[n][0][0]) + numbering[n][0][1])
						elif chain == "L":
							if frame != "cdrl3":    
								quality = ("bad", str(numbering[n][0][0]) + numbering[n][0][1])						
					else:
						all_region_aa[frame][str(numbering[n][0][0])] = numbering[n][1]
						if chain == "L":
							if frame == "fwl1":
								if str(numbering[n][0][0]) == "23":
									Conserved_position += 1
							if frame == "fwl3":
								if str(numbering[n][0][0]) == "104":
									Conserved_position += 1
								if str(numbering[n][0][0]) == "81" or str(numbering[n][0][0]) == "82":
									absent_positions -= 1
								elif str(numbering[n][0][0]) == "73":
									position_73 += 1
							elif frame == "fwl4":
								if str(numbering[n][0][0]) == "118":
									Conserved_position += 1
						elif chain == "H":
							if frame == "fwh1":
								if str(numbering[n][0][0]) == "23":
									Conserved_position += 1
								elif str(numbering[n][0][0]) == "2" and species == "rabbit":
									rabbit_position_2 -=1
							if frame == "fwh3":
								if str(numbering[n][0][0]) == "104":
									Conserved_position += 1
								elif str(numbering[n][0][0]) == "73":
									position_73 += 1
								elif str(numbering[n][0][0]) == "84" and species == "rabbit":
									rabbit_position_84 -=1
							elif frame == "fwh4":
								if str(numbering[n][0][0]) == "118":
									Conserved_position += 1
					region += numbering[n][1]
		regions[frame] = region   # Here we collect amino acid sequences for every antibody region
	if Conserved_position != 3:
		quality = ("Bad", "No Conserved Position")
		return all_region_aa, quality, regions
	if chain == "H" and species != "rabbit":
		quality = standard_species_length_check(position_73,regions,quality)
	elif chain == "H" and species == "rabbit":
		quality = rabbit_length_check(position_73,rabbit_position_84,regions,quality)
	elif chain == "L":
		quality = light_chain_species_length_check(absent_positions,position_73,regions,quality)
	if quality == "good":   
		######################################################################################
		### Check for indels in framework 1 as some sequences do not start with position 1 ###
		######################################################################################  
		if chain == "H":
			fw = "fwh1"
		else:
			fw = "fwl1"
		if species != "rabbit":
			quality = standard_species_length_check_fw1(fw,all_region_aa,quality)
		##################################################
		########### Special case for Rabbits #############
		##################################################
		else:
			quality = rabbit_length_check_fw1(fw,all_region_aa,rabbit_position_2,quality)
	return all_region_aa, quality, regions

def anarci_fun(list_of_inputs, ncores):
	"""
	Calling ANARCI here. We upload multiple sequences to ANARCI
	"""
	list_for_anarci_parsing = [("scFv"+str(n), list_of_inputs[n]) for n in range(len(list_of_inputs))]
	try:
		output = run_anarci(list_for_anarci_parsing, scheme='imgt', assign_germline = True, ncpu= ncores)
	except:
		print "ANARCI parsing failed to parse this chunk of sequences"			
	return output

def remove_stars(lis):
	return [x for x in lis if "*" not in x and len(x) < 152]  # We put length cut-off here


def check_V_and_J_gene_IMG_alignment(seq_id,species_matrix,gene):
	"""
	Here I check that V and J genes align to IMGT germline with
	at least 50% sequence identity
	"""
	IG_gene = None
	if species_matrix[seq_id][0]["germlines"][gene][0][0][0] != species:
		range_genes = len(species_matrix[seq_id][0]["germlines"][gene][0])
		for n in range(range_genes):
			if species_matrix[seq_id][0]["germlines"][gene][0][n][0] == species:
				if species_matrix[seq_id][0]["germlines"][gene][1][n] > 0.5:
					IG_gene = species_matrix[seq_id][0]["germlines"][gene][0][n][1]
				break
			else:
				if n == range_genes-1:
					IG_gene = None
	else:
		if species_matrix[seq_id][0]["germlines"][gene][1][0] > 0.50:  
			IG_gene = species_matrix[seq_id][0]["germlines"][gene][0][0][1]
		else:
			IG_gene = None
	return IG_gene

def check_IMGT_alignment(seq_id, species_matrix):
	#########################################
	### Here we aling sequence to HMM #######
	#########################################
	IGH_IGJ_status = "unknown"
	IG_V = check_V_and_J_gene_IMG_alignment(seq_id,species_matrix,"v_gene")
	IG_J = check_V_and_J_gene_IMG_alignment(seq_id,species_matrix,"j_gene")
	if chain == "H":	
		if not IG_J or not IG_V:
			IGH_IGJ_status = None
		elif IG_V == None or IG_J == None:
			IGH_IGJ_status = None
		elif IG_V[:3] != "IGH" or IG_J[:3] != "IGH":
			IGH_IGJ_status = None
	elif chain == "L":
		if not IG_J or not IG_V:
			IGH_IGJ_status = None
		elif IG_V[:3] == "IGH" or IG_J[:3] == "IGH":
			IGH_IGJ_status = None
		elif IG_V == None or IG_J == None:
			IGH_IGJ_status = None
	return IG_V, IG_J, IGH_IGJ_status

def checking_sequence_viability(output, species_matrix, inputs,SequenceTracker):
	sequence_correct = {}
	correct_V_gene = {}
	correct_J_gene = {}
	tuple_to_csvs = []
#	tuple_to_csv = namedtuple("tuple_to_csv", "full CDR3 V J AA Track")
	for u in range(len(output)):
		if not output[u] or output[u][0][0] == []:
			continue
		else:
			numbering = output[u][0][0]
			aa, status, regions = checking_structural_viability(numbering)
			if status != "good":
				continue
			if len(inputs[u][inputs[u].rfind(regions[list_of_frames[-1]]):]) > 13 or inputs[u].count(regions[list_of_frames[-1]]) > 1:   # Here we check for sneaky chimeric antibody sequences
				continue
			####################################################
			### Here we check for the correct antibody chain ###
			####################################################
			IG_V, IG_J, IGH_IGJ_status = check_IMGT_alignment(u,species_matrix)
			if IGH_IGJ_status == None:
				continue
			sequence_correct[inputs[u]] = json.dumps(aa)
			correct_V_gene[IG_V] = 1
			correct_J_gene[IG_J] = 1
			################################################################################
			### Here I write a temporary CSV to save CDR-H3, numbered antibody sequences ###
			################################################################################
			if chain == "H":
				tuple_to_csvs.append((inputs[u], regions["cdrh3"],IG_V,IG_J,json.dumps(aa), SequenceTracker.dictionary[inputs[u]]))

			else:
				tuple_to_csvs.append((inputs[u], regions["cdrl3"],IG_V,IG_J,json.dumps(aa), SequenceTracker.dictionary[inputs[u]]))

				
	return sequence_correct, correct_V_gene, correct_J_gene,tuple_to_csvs

def write_to_tempcsv(csv_list):
	for entry in csv_list:
		cwd = os.getcwd()
	
		with open(join(cwd,"Temp_CSV","temporary.csv"), "ab") as csv_file:
			writer = csv.writer(csv_file, delimiter=',')
			row = entry[0], entry[1],entry[2],entry[3],entry[4], entry[5]
			writer.writerow(row)

class TrackSequences():
	
	def __init__(self):

		self.dictionary = {}
		
	def add_entry(self,aa,query):
		if type(query) == int:
			self.dictionary[aa] = "Missing"		
		else:
			self.dictionary[aa] = query

	def __len__(self):

		return len(self.dictionary)

	def reset(self):

		self.dictionary = {}


def main(lines,ncpu,rate_of_analysis,frames, Antibodychain, SPECies):
	global list_of_frames
	global species
	global chain
	global SequenceTracker
	
	chain = Antibodychain
	species = SPECies
	list_of_frames = frames
	chunk_size = 500000

	t1 = time.time()
	correct_sequences = {}
	all_regions_aa = {}
	correct_V_gene = {}
	correct_J_gene = {}
	cwd = os.getcwd()
	
	try:
		os.makedirs(join(cwd,"Temp_CSV"))
	except:
		with open(join(cwd,"Temp_CSV","temporary.csv"), "wb") as csv_file:
			writer = csv.writer(csv_file, delimiter=',')
	################################
	### Need to add V,J gene dict ##
	################################
	inputs = []
	count = 0
	count_correct = 0
	num_of_entries = len(lines)
	print "The total number of entries is %d" %(num_of_entries)
	try:
		print "The total number of redundant reads is %d" % (sum([ len(lines[ss]) for ss in lines  ]))
	except TypeError:
		print "The total number of redundant reads is %d" % (sum(lines.values()))
	for frame in list_of_frames:
		if frame not in all_regions_aa:
			all_regions_aa[frame] = {}
	SequenceTracker = TrackSequences()
	for line in lines:
		if num_of_entries - count <= rate_of_analysis:
			rate_of_analysis = num_of_entries - count
			if rate_of_analysis < 1:
				continue
		SequenceTracker.add_entry(line, lines[line])


		if len(SequenceTracker)>= rate_of_analysis: 
			count += rate_of_analysis
			print count 
#			if count >=10001:
#				break
			
			inputs = remove_stars(SequenceTracker.dictionary.keys())
			output = anarci_fun(inputs, ncpu)
			##############################################################
			#### Here we parallelize ABOSS and analyse ABOSS outputs #####
			##############################################################

			pool = mp.Pool(processes=ncpu)
			results = [ pool.apply_async(checking_sequence_viability, args = (output[1][i::ncpu], output[2][i::ncpu], inputs[i::ncpu],SequenceTracker)) for i in xrange(ncpu) ]
			correct_sequences_portion = [ r.get() for r in results ]
			del results
			inputs = []
			SequenceTracker.reset()
			pool.close()
			temp_write_to_csv_list = []
			for portion in range(len(correct_sequences_portion)):				
				correct_sequences = merge_dicts(correct_sequences_portion[portion][0], correct_sequences)
				correct_V_gene = merge_dicts(correct_sequences_portion[portion][1], correct_V_gene)
				correct_J_gene = merge_dicts(correct_sequences_portion[portion][2], correct_J_gene)
				temp_write_to_csv_list += correct_sequences_portion[portion][3]
			
			"""
			write output to temp file
			"""
			write_to_tempcsv(temp_write_to_csv_list)
			temp_write_to_csv_list = []


			del correct_sequences_portion
			print len(correct_sequences) + count_correct
			print "Time took ", time.time()-t1

			##########################################################
			###  Here I save ANARCI parsed sequences to JSON files ###
			##########################################################
			if len(correct_sequences) >= chunk_size:   
				for sequence in correct_sequences:
					if type(lines[sequence]) == int:
						redundancy = lines[sequence]
					else:
						redundancy = len(lines[sequence])
					aa = json.loads(correct_sequences[sequence])
					for framework in aa:	   
						for pos in aa[framework]:
							for ele in aa[framework][pos]:
								if pos not in all_regions_aa[framework]:
									all_regions_aa[framework][pos] = {}
								if ele in all_regions_aa[framework][pos]:
									all_regions_aa[framework][pos][ele] += redundancy
								elif ele not in all_regions_aa[framework][pos]:
									all_regions_aa[framework][pos][ele] = redundancy
				if chain == "L":
					print all_regions_aa["fwl3"]["104"]
					print all_regions_aa["fwl1"]["23"]
				if chain == "H":
					print all_regions_aa["fwh3"]["104"]
					print all_regions_aa["fwh1"]["23"]
				count_correct += len(correct_sequences)
				correct_sequences = {}
			##########################################################
			### Here we create amino acid distribution by position ###
			##########################################################		
			print "The proportion of sequences passed the filtering step %f" % (round((float(len(correct_sequences) + count_correct))/count,3))
			print "---------------------------------"

	#################################
	# Here we save all passed reads #
	#################################
	for sequence in correct_sequences:
		if type(lines[sequence]) == int:
			redundancy = lines[sequence]
		else:
			redundancy = len(lines[sequence])
		aa = json.loads(correct_sequences[sequence])
		for framework in aa:	   
			for pos in aa[framework]:
				for ele in aa[framework][pos]:
					if pos not in all_regions_aa[framework]:
						all_regions_aa[framework][pos] = {}
					if ele in all_regions_aa[framework][pos]:
						all_regions_aa[framework][pos][ele] += redundancy
					elif ele not in all_regions_aa[framework][pos]:
						all_regions_aa[framework][pos][ele] = redundancy
	correct_sequences = {}
	del lines
	with open("counting_aa_raw.pkl", "wb") as s:    # Here we save amino acid propotions by position
		pickle.dump(all_regions_aa, s, protocol=pickle.HIGHEST_PROTOCOL)
	with open("correct_V_gene.pkl", "wb") as s:    # Here we save V genes
		pickle.dump(correct_V_gene, s, protocol=pickle.HIGHEST_PROTOCOL)
	with open("correct_J_gene.pkl", "wb") as s:    # Here we save J genes
		pickle.dump(correct_J_gene, s, protocol=pickle.HIGHEST_PROTOCOL)
	if chain == "L":
		print all_regions_aa["fwl3"]["104"]
		print all_regions_aa["fwl1"]["23"]
	if chain == "H":
		print all_regions_aa["fwh3"]["104"]
		print all_regions_aa["fwh1"]["23"]
	return count_correct
 
if __name__ == '__main__':
	pass
